###############################################################################
# Backward solution for life cycle planning problem                           #
###############################################################################


#### reset AMPL ###############################################################
reset;


#### define solver ############################################################

option solver knitroampl;
option knitro_options "outlev=0";
     

#### constants ################################################################

param pi := 3.141592654;


#### definition of chebychev polynomials ######################################

param M;	                                        # number of nodes;
param lb;                                           # lower state bound;
param ub;                                           # upper state bound;
set nodes := 1..M;                                  # set of nodes


#### definition of parameters #################################################

param gamma;                                        # relative risk aversion
param T;                                            # number of periods
set time := 1..T;                                   # set of times
param beta;                                         # time discount factor
param r;                                            # interest rate
                                                 
param ns;                                           # number of states
set states := 1..ns;                                # set of states
param prob{i in states, j in states} default 1/ns;  # transition probabilities


#### income process ###########################################################

param y{s in states, t in time} default 0;          # income in time and state


#### stock process ############################################################

param R{s in states, t in time} default 0;          # stock return process                                                         
                      

#### counter and other variables ##############################################

param t_count;                                      # age counter


#### set up data ############################################################## 

param borrowsafe default 0;
param borrowrisk default 0;
param optlb{t in time} default 0;

#### change data input file here
data projectdata_1.dat  
 
param Rmin := min{s in states, t in time} R[s,t];   # minimum stock return
param Rmax := max{s in states, t in time} R[s,t];   # maximum stock return
let {t in time: t <= 6} optlb[t] := lb;
 

#### set up nodes and interpolation coefficients ##############################

param z{k in nodes} := -cos((2*k-1)*pi/(2*M));      # nodes on [-1,1]
param x{k in nodes} := (z[k]+1)*(ub-lb)/2+lb;       # adjusted nodes to [lb,ub]
param data{s in states, k in nodes, t in time}      # interpolation data
      default 0;     
param phi{k in nodes, i in nodes} :=                # collocation matrix
      cos((k-1)*acos(z[i]));                        
param coeff{s in states, i in nodes, t in time} :=  # interp. coefficients
      (sum{k in nodes}(data[s,k,t]*phi[i,k]))/
      (sum{k in nodes}(phi[i,k]^2));                


#### policy variables #########################################################

param c{s in states, k in nodes, t in time}         # consumption policy
      default 0;         
param Wsafe{s in states, k in nodes, t in time}     # bond policy
      default 0;         
param Wrisk{s in states, k in nodes, t in time}     # stock policy
      default 0;
                                                                  

#### optimization problem #####################################################

#### optimization variable
var assets{s in states, k in nodes} >= borrowsafe;
var risky{s in states, k in nodes} >= borrowrisk;

#### optimization problem
maximize value: sum{s in states, k in nodes}        # objective function
      (((x[k]+y[s,t_count]-assets[s,k]-risky[s,k])^(1-gamma))+    
      beta*(sum{j in states}prob[j,s]*                            
      (sum{i in nodes} coeff[j,i,t_count+1]*cos((i-1)*
      acos(2*(risky[s,k]*R[j,t_count+1]+assets[s,k]*(1+r)-lb)/(ub-lb) - 1))
      )^(1-gamma)))^(1/(1-gamma));

#### constraints
subject to lowb {s in states,k in nodes}:           # lower bound on grid
      assets[s,k]*(1+r)+risky[s,k]*Rmin >= lb;      
subject to upb {s in states,k in nodes}:            # upper bound on grid
      assets[s,k]*(1+r) + risky[s,k]*Rmax <= ub;
subject to cons {s in states,k in nodes}:           # consumtion nonnegative
      x[k]+y[s,t_count]-assets[s,k]-risky[s,k] >= 0 ;

#### problem definition
problem Bellman: value, assets, risky, lowb, upb, cons;


#### set data for last period #################################################

let {s in states, k in nodes} data[s,k,T] := x[k]+y[s,T];
let {s in states, k in nodes} c[s,k,T] := x[k]+y[s,T];
let t_count := T-1;
    
printf "\n\n";
            

#### do-loop for backward optimization ########################################
repeat{

    printf "Starting solution at age %2d:\n\n", t_count;

    let {s in states, k in nodes} assets[s, k] := 0;
    let {s in states, k in nodes} risky[s, k] := 0;

    #### solver problem
    solve Bellman;

    #### store new values
    let {s in states, k in nodes} c[s,k,t_count] := 
          x[k]+y[s,t_count]-assets[s,k]-risky[s,k];
    let {s in states, k in nodes} Wsafe[s,k,t_count] := assets[s,k];
    let {s in states, k in nodes} Wrisk[s,k,t_count] := risky[s,k];
    let {s in states, k in nodes} data[s,k, t_count] := 
          ((c[s,k,t_count]^(1-gamma)) +
          beta*(sum{j in states}prob[j,s]*(sum{i in nodes} 
          coeff[j,i,t_count+1]*cos((i-1)*
          acos(2*(risky[s,k]*R[j,t_count+1]+assets[s,k]*(1+r)
          -lb)/(ub-lb) - 1)))^(1-gamma)))^(1/(1-gamma));
    let t_count := t_count - 1;   
    
    printf "\n\n"; 
} while ( t_count > 0);


#### interpolate asset policies ###############################################
      
param coeffsafe{s in states, i in nodes, t in time} :=  # interp. coefficients
      (sum{k in nodes}(Wsafe[s,k,t]*phi[i,k]))/                            
      (sum{k in nodes}(phi[i,k]^2));                                      
param coeffrisk{s in states, i in nodes, t in time} :=  # interp. coefficients
      (sum{k in nodes}(Wrisk[s,k,t]*phi[i,k]))/                            
      (sum{k in nodes}(phi[i,k]^2));


#### output data to datapolicy.m ##############################################

#### parameters
printf "T = %6d;\", T > datapolicy.m;
printf "ns = %6d;\n", ns > datapolicy.m;
printf "M = %6d;\n", M > datapolicy.m;
printf "lb = %25.16e;\n", lb  > datapolicy.m;
printf "ub = %25.16e;\n", ub  > datapolicy.m;
printf "r = %25.16e;\n", r  > datapolicy.m;

#### probabilities
printf "prob = zeros(%6d, %6d);\n", ns, ns > datapolicy.m;
for{s in states, t in states}{
    printf "prob(%6d, %6d) = %25.16e;\n", s, t, prob[s, t]  > datapolicy.m;
}

#### rates of return
printf "R = zeros(%6d, %6d);\n", ns, T > datapolicy.m;
for{s in states, t in time}{
    printf "R(%6d, %6d) = %25.16e;\n", s, t, R[s, t]  > datapolicy.m;
}

#### income process
printf "y = zeros(%6d, %6d);\n", ns, T > datapolicy.m;
for{s in states, t in time}{
    printf "y(%6d, %6d) = %25.16e;\n", s, t, y[s, t]  > datapolicy.m;
}

#### collocation nodes
printf "xt = zeros(%6d, 1);\n", M > datapolicy.m;
for{i in nodes}{
    printf "xt(%6d) = %25.16e;\n", i, x[i] > datapolicy.m;
}

#### polynomial coefficients
printf "coeffsafe = zeros(%2d, %6d, %2d);\n", ns, M, T > datapolicy.m;
printf "coeffrisk = zeros(%2d, %6d, %2d);\n", ns, M, T > datapolicy.m;
printf "coeffvalue = zeros(%2d, %6d, %2d);\n", ns, M, T > datapolicy.m;
for{s in states, i in nodes, t in time}{
    printf "coeffsafe(%2d, %6d, %2d) = %25.16e;\n", s, i, t, 
          coeffsafe[s, i, t]  > datapolicy.m;
    printf "coeffrisk(%2d, %6d, %2d) = %25.16e;\n", s, i, t, 
          coeffrisk[s, i, t]  > datapolicy.m;      
    printf "coeffvalue(%2d, %6d, %2d) = %25.16e;\n", s, i, t, 
          coeff[s, i, t]  > datapolicy.m;
}
